/*
 * libipcountry.c:
 * Library interface to IP-to-country database.
 *
 * Copyright (c) 2003 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: libipcountry.c,v 1.1 2003/12/06 16:04:35 chris Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include "ipcountry.h"

#ifndef DATAFILE
#   define DATAFILE    "/software/share/ip-to-country/ip-to-country.dat"
#endif

#define RECORDLEN   10      /* 2 * sizeof(unit32_t) + 2 * sizeof(char) */

/* struct ipcountry
 * Implementation of database object. */
struct ipcountry {
    FILE *ic_fp;
    long ic_nrecs;
};

/* read_rec STREAM INDEX LOW HIGH CODE
 * Read a record from STREAM. */
static int read_rec(FILE *fp, const long index, uint32_t *lo, uint32_t *hi, char *cc) {
    char buf[10];
    if (-1 == fseek(fp, index * RECORDLEN, SEEK_SET)
        || 1 != fread(buf, RECORDLEN, 1, fp))
        return -1;
    memcpy(lo, buf, sizeof *lo);
    memcpy(hi, buf + 4, sizeof *hi);
    if (cc) {
        memcpy(cc, buf + 8, 2);
        cc[2] = 0;
    }
    return 0;
}

/* ipcountry_open DATABASE
 * Open ipcountry DATABASE for queries (or the default location if DATABASE is
 * NULL). Returns a database handle on success or false on failure. */
ipcountry ipcountry_open(const char *dbname) {
    ipcountry I;
    FILE *fp;
    const struct ipcountry Iz = {0};
    long nrecs;

    if (!(fp = fopen(dbname ? dbname : DATAFILE, "rb"))
        || -1 == fseek(fp, 0, SEEK_END)
        || -1 == (nrecs = ftell(fp) / RECORDLEN)
        || -1 == fseek(fp, 0, SEEK_SET)) {
        if (fp) fclose(fp);
        return NULL;
    }

    if (!(I = malloc(sizeof *I))) {
        fclose(fp);
        return NULL;
    }
    
    *I = Iz;
    I->ic_fp = fp;
    I->ic_nrecs = nrecs;
    return I;
}

/* ipcountry_close DB
 * Close DB. */
void ipcountry_close(ipcountry I) {
    if (I) {
        if (I->ic_fp) fclose(I->ic_fp);
        free(I);
    }
}

/* ipcountry_find DB ADDR CODE
 * Find ADDR in the IP-to-country DB, and saving it in CODE, which must point
 * to at least three bytes of memory. Returns ipcr_ok if a country was found,
 * ipcr_not_found if it was not, and ipcr_fatal in case of an unrecoverable
 * error which leaves the DB handle unusable. */
enum ipcountry_result ipcountry_find(ipcountry I, const struct in_addr in, char *cc) {
#   define fp  (I->ic_fp)
    uint32_t olo, ohi, ilo, ihi, n;
    
    if (!I)
        return ipcr_fatal;
    
    olo = 0;
    ohi = I->ic_nrecs - 1;
    
    n = (uint32_t)ntohl(in.s_addr);

    do {
        long o;

        o = (ohi + olo) / 2;
        if (-1 == read_rec(fp, o, &ilo, &ihi, cc))
            return ipcr_fatal;
        
        if (n < ilo)
            ohi = o;
        else if (n > ihi)
            olo = o;
        else
            break;
    } while (ohi > olo + 1);

    for (; olo <= ohi; ++olo) {
        if (-1 == read_rec(fp, olo, &ilo, &ihi, cc))
            return ipcr_fatal;
        else if (n >= ilo && n <= ihi)
            return ipcr_ok;
    }
    return ipcr_not_found;
#   undef fp
}


